from __future__ import annotations

import ast
import asyncio
import itertools
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Coroutine,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from ....utils.async_itertools import async_next
from ....utils.async_tools import create_sub_task, run_coroutine_in_thread, threaded
from ....utils.glob_path import iter_files
from ....utils.logging import LoggingDescriptor
from ....utils.uri import Uri
from ...common.decorators import language_id
from ...common.lsp_types import Location, Position, ReferenceContext
from ...common.text_document import TextDocument
from ..configuration import WorkspaceConfig
from ..diagnostics.entities import (
    ArgumentDefinition,
    LocalVariableDefinition,
    VariableDefinition,
)
from ..diagnostics.library_doc import (
    ALL_RUN_KEYWORDS_MATCHERS,
    RESOURCE_FILE_EXTENSION,
    ROBOT_FILE_EXTENSION,
    KeywordDoc,
    KeywordMatcher,
    LibraryDoc,
)
from ..diagnostics.namespace import Namespace
from ..utils.ast_utils import (
    HasTokens,
    Statement,
    Token,
    get_nodes_at_position,
    get_tokens_at_position,
    is_not_variable_token,
    iter_over_keyword_names_and_owners,
    range_from_token,
    tokenize_variables,
)
from ..utils.async_ast import iter_nodes

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart

_ReferencesMethod = Callable[[ast.AST, TextDocument, Position, ReferenceContext], Awaitable[Optional[List[Location]]]]


class RobotReferencesProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.references.collect.add(self.collect)

    def _find_method(self, cls: Type[Any]) -> Optional[_ReferencesMethod]:
        if cls is ast.AST:
            return None
        method_name = "references_" + cls.__name__
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            if callable(method):
                return cast(_ReferencesMethod, method)
        for base in cls.__bases__:
            method = self._find_method(base)
            if method:
                return cast(_ReferencesMethod, method)
        return None

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def collect(
        self, sender: Any, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:

        result_nodes = await get_nodes_at_position(await self.parent.documents_cache.get_model(document), position)

        if not result_nodes:
            return None

        result_node = result_nodes[-1]

        if result_node is None:
            return None

        result = await self._references_default(result_nodes, document, position, context)
        if result:
            return result

        method = self._find_method(type(result_node))
        if method is not None:
            result = await method(result_node, document, position, context)
            if result is not None:
                return result

        return None

    async def _find_references(
        self,
        document: TextDocument,
        func: Callable[..., Coroutine[None, None, List[Location]]],
        *args: Any,
        **kwargs: Any,
    ) -> List[Location]:
        folder = self.parent.workspace.get_workspace_folder(document.uri)
        if folder is None:
            return []

        futures: List[Awaitable[List[Location]]] = []
        result: List[Location] = []

        config = await self.parent.workspace.get_configuration(WorkspaceConfig, folder.uri) or WorkspaceConfig()

        async for f in iter_files(
            folder.uri.to_path(),
            (f"**/*.{{{ROBOT_FILE_EXTENSION[1:]},{RESOURCE_FILE_EXTENSION[1:]}}}"),
            ignore_patterns=config.exclude_patterns or [],  # type: ignore
            absolute=True,
        ):
            try:
                doc = await self.parent.robot_workspace.get_or_open_document(f, "robotframework")
            except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
                raise
            except BaseException as ex:
                self._logger.exception(ex)
            else:
                futures.append(run_coroutine_in_thread(func, doc, *args, **kwargs))

        for e in await asyncio.gather(*futures, return_exceptions=True):
            if isinstance(e, BaseException):
                if not isinstance(result, asyncio.CancelledError):
                    self._logger.exception(e)
                continue
            result.extend(e)

        return result

    async def _references_default(
        self, nodes: List[ast.AST], document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        if not nodes:
            return None

        node = nodes[-1]

        if not isinstance(node, HasTokens):
            return None

        tokens = get_tokens_at_position(node, position)

        token_and_var: Optional[Tuple[Token, VariableDefinition]] = None

        for token in tokens:
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_variables_from_token(token, namespace, nodes, position)
                    if position in range_from_token(var_token)
                ),
                None,
            )

        if (
            token_and_var is None
            and isinstance(node, Statement)
            and isinstance(node, self.get_expression_statement_types())
            and (token := node.get_token(RobotToken.ARGUMENT)) is not None
            and position in range_from_token(token)
        ):
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_expression_variables_from_token(
                        token, namespace, nodes, position
                    )
                    if position in range_from_token(var_token)
                ),
                None,
            )

        if token_and_var is not None:
            _, variable = token_and_var

            return [
                *(
                    [
                        Location(
                            uri=str(Uri.from_path(variable.source)),
                            range=range_from_token(variable.name_token) if variable.name_token else variable.range(),
                        ),
                    ]
                    if context.include_declaration and variable.source
                    else []
                ),
                *(await self.find_variable_references(document, variable)),
            ]

        return None

    async def find_variable_references(self, document: TextDocument, variable: VariableDefinition) -> List[Location]:
        return (
            await create_sub_task(self.find_variable_references_in_file(document, variable))
            if isinstance(variable, (ArgumentDefinition, LocalVariableDefinition))
            else await self._find_references(document, self.find_variable_references_in_file, variable)
        )

    async def yield_argument_name_and_rest(self, node: ast.AST, token: Token) -> AsyncGenerator[Token, None]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Arguments

        if isinstance(node, Arguments) and token.type == RobotToken.ARGUMENT:
            argument = next(
                (
                    v
                    for v in itertools.dropwhile(
                        lambda t: t.type in RobotToken.NON_DATA_TOKENS,
                        tokenize_variables(token, ignore_errors=True),
                    )
                    if v.type == RobotToken.VARIABLE
                ),
                None,
            )
            if argument is None or argument.value == token.value:
                yield token
            else:
                yield argument
                i = len(argument.value)

                async for t in self.yield_argument_name_and_rest(
                    node, RobotToken(token.type, token.value[i:], token.lineno, token.col_offset + i, token.error)
                ):
                    yield t
        else:
            yield token

    @_logger.call
    async def find_variable_references_in_file(
        self,
        doc: TextDocument,
        variable: VariableDefinition,
    ) -> List[Location]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.blocks import Block, Keyword, Section, TestCase

        namespace = await self.parent.documents_cache.get_namespace(doc)
        model = await self.parent.documents_cache.get_model(doc)

        if (
            variable.source
            and variable.source != str(doc.uri.to_path())
            and not any(
                e for e in (await namespace.get_resources()).values() if e.library_doc.source == variable.source
            )
            and not any(
                e
                for e in (await namespace.get_imported_variables()).values()
                if e.library_doc.source == variable.source
            )
            and not any(e for e in await namespace.get_command_line_variables() if e.source == variable.source)
        ):

            return []

        expression_statements = self.get_expression_statement_types()

        result: List[Location] = []
        current_block: Optional[Block] = None

        async for node in iter_nodes(model):
            if isinstance(node, Section):
                current_block = None
            elif isinstance(node, (TestCase, Keyword)):
                current_block = node

            if isinstance(node, HasTokens):
                for token1 in node.tokens:
                    async for token in self.yield_argument_name_and_rest(node, token1):
                        async for token_and_var in self.iter_variables_from_token(
                            token,
                            namespace,
                            [*([current_block] if current_block is not None else []), node],
                            range_from_token(token).start,
                        ):
                            sub_token, found_variable = token_and_var

                            if found_variable == variable:
                                result.append(Location(str(doc.uri), range_from_token(sub_token)))

            if (
                isinstance(node, Statement)
                and isinstance(node, expression_statements)
                and (token := node.get_token(RobotToken.ARGUMENT)) is not None
            ):
                async for token_and_var in self.iter_expression_variables_from_token(
                    token,
                    namespace,
                    [*([current_block] if current_block is not None else []), node],
                    range_from_token(token).start,
                ):
                    sub_token, found_variable = token_and_var

                    if found_variable == variable:
                        result.append(Location(str(doc.uri), range_from_token(sub_token)))

        return result

    async def references_KeywordName(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordName

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordName, node)

        name_token = cast(RobotToken, kw_node.get_token(RobotToken.KEYWORD_NAME))

        if not name_token:
            return None

        doc = await namespace.get_library_doc()
        if doc is not None:
            keyword = next(
                (v for v in doc.keywords.keywords if v.name == name_token.value and v.line_no == kw_node.lineno),
                None,
            )

            if keyword is not None and keyword.source and not keyword.is_error_handler:
                return [
                    *(
                        [
                            Location(
                                uri=str(Uri.from_path(keyword.source)),
                                range=keyword.range,
                            )
                        ]
                        if context.include_declaration and keyword.source
                        else []
                    ),
                    *await self.find_keyword_references(document, keyword),
                ]

        return None

    async def references_KeywordCall(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordCall

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordCall, node)
        result = await self.get_keyworddoc_and_token_from_position(
            kw_node.keyword,
            cast(Token, kw_node.get_token(RobotToken.KEYWORD)),
            [cast(Token, t) for t in kw_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    # TODO: find references for Library Namespace
                    return None

            if position in kw_range and keyword_doc is not None:
                return [
                    *(
                        [Location(uri=str(Uri.from_path(keyword_doc.source)), range=keyword_doc.range)]
                        if context.include_declaration and keyword_doc.source
                        else []
                    ),
                    *await self.find_keyword_references(document, keyword_doc),
                ]

        return None

    async def references_Fixture(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Fixture

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        fixture_node = cast(Fixture, node)

        name_token = cast(Token, fixture_node.get_token(RobotToken.NAME))
        if name_token is None or name_token.value is None or name_token.value.upper() in ("", "NONE"):
            return None

        result = await self.get_keyworddoc_and_token_from_position(
            fixture_node.name,
            name_token,
            [cast(Token, t) for t in fixture_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    # TODO: find references for Library Namespace
                    return None

            if position in kw_range and keyword_doc is not None and not keyword_doc.is_error_handler:
                return [
                    *(
                        [Location(uri=str(Uri.from_path(keyword_doc.source)), range=keyword_doc.range)]
                        if context.include_declaration and keyword_doc.source
                        else []
                    ),
                    *await self.find_keyword_references(document, keyword_doc),
                ]

        return None

    async def _references_Template_or_TestTemplate(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Template, TestTemplate

        node = cast(Union[Template, TestTemplate], node)
        if node.value:

            keyword_token = cast(RobotToken, node.get_token(RobotToken.NAME, RobotToken.ARGUMENT))
            if keyword_token is None or keyword_token.value is None or keyword_token.value.upper() in ("", "NONE"):
                return None

            keyword_token = self.strip_bdd_prefix(keyword_token)

            if position.is_in_range(range_from_token(keyword_token)):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                keyword_doc = await namespace.find_keyword(node.value)

                if keyword_doc is not None:

                    lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

                    kw_range = range_from_token(keyword_token)

                    if lib_entry and kw_namespace:
                        r = range_from_token(keyword_token)
                        r.end.character = r.start.character + len(kw_namespace)
                        kw_range.start.character = r.end.character + 1
                        if position in r:
                            # TODO: find references for Library Namespace
                            return None
                    if keyword_doc:
                        return [
                            *(
                                [
                                    Location(
                                        uri=str(Uri.from_path(keyword_doc.source)),
                                        range=keyword_doc.range,
                                    )
                                ]
                                if context.include_declaration and keyword_doc.source
                                else []
                            ),
                            *await self.find_keyword_references(document, keyword_doc),
                        ]

        return None

    async def references_TestTemplate(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        return await self._references_Template_or_TestTemplate(node, document, position, context)

    async def references_Template(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        return await self._references_Template_or_TestTemplate(node, document, position, context)

    @_logger.call
    async def find_keyword_references_in_file(
        self,
        doc: TextDocument,
        kw_doc: KeywordDoc,
        lib_doc: Optional[LibraryDoc] = None,
    ) -> List[Location]:

        try:
            namespace = await self.parent.documents_cache.get_namespace(doc)

            if (
                lib_doc is not None
                and lib_doc.source is not None
                and lib_doc.source != str(doc.uri.to_path())
                and lib_doc not in (e.library_doc for e in (await namespace.get_libraries()).values())
                and lib_doc not in (e.library_doc for e in (await namespace.get_resources()).values())
            ):
                return []

            return await self._find_keyword_references_in_namespace(namespace, kw_doc)
        except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
            raise
        except BaseException as e:
            self._logger.exception(e)

        return []

    async def _find_keyword_references_in_namespace(self, namespace: Namespace, kw_doc: KeywordDoc) -> List[Location]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import (
            Fixture,
            KeywordCall,
            Template,
            TestTemplate,
        )

        result: List[Location] = []

        async for node in iter_nodes(namespace.model):

            kw_token: Optional[Token] = None
            arguments: Optional[List[Token]] = None

            if isinstance(node, KeywordCall):
                kw_token = node.get_token(RobotToken.KEYWORD)
                arguments = list(node.get_tokens(RobotToken.ARGUMENT) or [])

                async for location in self.get_keyword_references_from_tokens(
                    namespace, kw_doc, node, kw_token, arguments
                ):
                    result.append(location)
            elif isinstance(node, Fixture):
                kw_token = node.get_token(RobotToken.NAME)
                arguments = list(node.get_tokens(RobotToken.ARGUMENT) or [])

                async for location in self.get_keyword_references_from_tokens(
                    namespace, kw_doc, node, kw_token, arguments
                ):
                    result.append(location)
            elif isinstance(node, (Template, TestTemplate)):
                kw_token = node.get_token(RobotToken.NAME, RobotToken.ARGUMENT)

                async for location in self.get_keyword_references_from_tokens(namespace, kw_doc, node, kw_token, []):
                    result.append(location)

        return result

    async def get_keyword_references_from_tokens(
        self,
        namespace: Namespace,
        kw_doc: KeywordDoc,
        node: ast.AST,
        kw_token: Optional[Token],
        arguments: Optional[List[Token]],
        unescape_kw_token: bool = False,
    ) -> AsyncGenerator[Location, None]:
        from robot.utils.escaping import unescape

        if kw_token is not None and is_not_variable_token(kw_token):
            kw_token = self.strip_bdd_prefix(kw_token)

            kw: Optional[KeywordDoc] = None
            kw_matcher = KeywordMatcher(kw_doc.name)
            kw_name = unescape(kw_token.value) if unescape_kw_token else kw_token.value

            libraries_matchers = await namespace.get_libraries_matchers()
            resources_matchers = await namespace.get_resources_matchers()

            for lib, name in iter_over_keyword_names_and_owners(kw_name):
                if (
                    lib is not None
                    and not any(k for k in libraries_matchers.keys() if k == lib)
                    and not any(k for k in resources_matchers.keys() if k == lib)
                ):
                    continue

                lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, kw_token)
                kw_range = range_from_token(kw_token)

                if lib_entry and kw_namespace:
                    r = range_from_token(kw_token)
                    r.end.character = r.start.character + len(kw_namespace)
                    kw_range.start.character = r.end.character + 1

                if name is not None:
                    if kw_matcher == name:
                        kw = await namespace.find_keyword(kw_name)

                        if kw is not None and kw == kw_doc:
                            yield Location(
                                str(Uri.from_path(namespace.source).normalized()),
                                range=kw_range,
                            )

                    if any(k for k in ALL_RUN_KEYWORDS_MATCHERS if k == name) and arguments:
                        async for location in self.get_keyword_references_from_any_run_keyword(
                            namespace, kw_doc, node, kw_token, arguments
                        ):
                            yield location

    async def get_keyword_references_from_any_run_keyword(
        self,
        namespace: Namespace,
        kw_doc: KeywordDoc,
        node: ast.AST,
        kw_token: Token,
        arguments: List[Token],
    ) -> AsyncGenerator[Location, None]:
        if kw_token is None or not is_not_variable_token(kw_token):
            return

        kw = await namespace.find_keyword(str(kw_token.value))

        if kw is None or not kw.is_any_run_keyword():
            return

        if kw.is_run_keyword() and len(arguments) > 0 and is_not_variable_token(arguments[0]):
            async for e in self.get_keyword_references_from_tokens(
                namespace, kw_doc, node, arguments[0], arguments[1:]
            ):
                yield e
        elif (
            kw.is_run_keyword_with_condition()
            and len(arguments) > (cond_count := kw.run_keyword_condition_count())
            and is_not_variable_token(arguments[1])
        ):
            async for e in self.get_keyword_references_from_tokens(
                namespace, kw_doc, node, arguments[cond_count], arguments[cond_count + 1 :], True
            ):
                yield e
        elif kw.is_run_keywords():
            has_separator = False
            while arguments:

                t = arguments[0]
                arguments = arguments[1:]
                if t.value == "AND":
                    continue

                separator_token = next((e for e in arguments if e.value == "AND"), None)
                args = []
                if separator_token is not None:
                    args = arguments[: arguments.index(separator_token)]
                    arguments = arguments[arguments.index(separator_token) + 1 :]
                    has_separator = True
                else:
                    if has_separator:
                        args = arguments

                if is_not_variable_token(t):
                    async for e in self.get_keyword_references_from_tokens(namespace, kw_doc, node, t, args, True):
                        yield e
        elif kw.is_run_keyword_if() and len(arguments) > 1:
            arguments = arguments[1:]

            while arguments:
                t = arguments[0]
                arguments = arguments[1:]

                if t.value in ["ELSE", "ELSE IF"]:
                    continue

                separator_token = next((e for e in arguments if e.value in ["ELSE", "ELSE IF"]), None)
                args = []
                if separator_token is not None:
                    args = arguments[: arguments.index(separator_token)]
                    arguments = arguments[arguments.index(separator_token) + 1 :]
                    if separator_token.value == "ELSE IF":
                        arguments = arguments[1:]

                if is_not_variable_token(t):
                    async for e in self.get_keyword_references_from_tokens(namespace, kw_doc, node, t, args, True):
                        yield e

    async def find_keyword_references(self, document: TextDocument, kw_doc: KeywordDoc) -> List[Location]:
        folder = self.parent.workspace.get_workspace_folder(document.uri)
        if folder is None:
            return []

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        lib_doc = (
            next(
                (
                    e.library_doc
                    for e in (await namespace.get_libraries()).values()
                    if kw_doc in e.library_doc.keywords.values()
                ),
                None,
            )
            or next(
                (
                    e.library_doc
                    for e in (await namespace.get_resources()).values()
                    if kw_doc in e.library_doc.keywords.values()
                ),
                None,
            )
            or await namespace.get_library_doc()
        )

        return await self._find_references(document, self.find_keyword_references_in_file, kw_doc, lib_doc)

    @_logger.call
    async def _find_library_import_references_in_file(
        self,
        doc: TextDocument,
        library_doc: LibraryDoc,
    ) -> List[Location]:

        namespace = await self.parent.documents_cache.get_namespace(doc)

        result: List[Location] = []
        for lib_entry in (await namespace.get_libraries()).values():
            if (
                lib_entry.import_source == str(doc.uri.to_path())
                and lib_entry.library_doc.source_or_origin == library_doc.source_or_origin
            ):
                result.append(Location(str(doc.uri), lib_entry.import_range))

        return result

    async def references_LibraryImport(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import LibraryImport

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        import_node = cast(LibraryImport, node)

        name_token = cast(RobotToken, import_node.get_token(RobotToken.NAME))

        if not name_token:
            return None

        if position in range_from_token(name_token):
            library_doc = await namespace.get_imported_library_libdoc(
                import_node.name, import_node.args, import_node.alias
            )

            if library_doc is None:
                return None

            return await self._find_references(document, self._find_library_import_references_in_file, library_doc)

        return None

    @_logger.call
    async def _find_resource_import_references_in_file(
        self,
        doc: TextDocument,
        library_doc: LibraryDoc,
    ) -> List[Location]:

        namespace = await self.parent.documents_cache.get_namespace(doc)

        result: List[Location] = []
        for lib_entry in (await namespace.get_resources()).values():
            if lib_entry.import_source == str(doc.uri.to_path()) and lib_entry.library_doc.source == library_doc.source:
                result.append(Location(str(doc.uri), lib_entry.import_range))

        return result

    async def references_ResourceImport(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import ResourceImport

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        import_node = cast(ResourceImport, node)

        name_token = cast(RobotToken, import_node.get_token(RobotToken.NAME))

        if not name_token:
            return None

        if position in range_from_token(name_token):
            library_doc = await namespace.get_imported_resource_libdoc(import_node.name)

            if library_doc is None:
                return None

            return await self._find_references(document, self._find_resource_import_references_in_file, library_doc)

        return None

    async def _find_variables_import_references_in_file(
        self,
        doc: TextDocument,
        library_doc: LibraryDoc,
    ) -> List[Location]:

        namespace = await self.parent.documents_cache.get_namespace(doc)

        result: List[Location] = []
        for lib_entry in (await namespace.get_imported_variables()).values():
            if lib_entry.import_source == str(doc.uri.to_path()) and lib_entry.library_doc.source == library_doc.source:
                result.append(Location(str(doc.uri), lib_entry.import_range))

        return result

    async def references_VariablesImport(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position, context: ReferenceContext
    ) -> Optional[List[Location]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import VariablesImport

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        import_node = cast(VariablesImport, node)

        name_token = cast(RobotToken, import_node.get_token(RobotToken.NAME))

        if not name_token:
            return None

        if position in range_from_token(name_token):
            library_doc = await namespace.get_imported_variables_libdoc(import_node.name, import_node.args)

            if library_doc is None:
                return None

            return await self._find_references(document, self._find_variables_import_references_in_file, library_doc)

        return None



from __future__ import annotations

from asyncio import CancelledError
from typing import TYPE_CHECKING, Any, List, Optional

from ....jsonrpc2.protocol import rpc_method
from ....utils.async_tools import async_tasking_event
from ....utils.logging import LoggingDescriptor
from ..decorators import language_id_filter
from ..has_extend_capabilities import HasExtendCapabilities
from ..lsp_types import (
    Position,
    SelectionRange,
    SelectionRangeOptions,
    SelectionRangeParams,
    ServerCapabilities,
    TextDocumentIdentifier,
)
from ..text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import LanguageServerProtocol

from .protocol_part import LanguageServerProtocolPart


class SelectionRangeProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):

    _logger = LoggingDescriptor()

    def __init__(self, parent: LanguageServerProtocol) -> None:
        super().__init__(parent)

    def extend_capabilities(self, capabilities: ServerCapabilities) -> None:
        if len(self.collect):
            capabilities.selection_range_provider = SelectionRangeOptions(work_done_progress=True)

    @async_tasking_event
    async def collect(
        sender, document: TextDocument, positions: List[Position]  # NOSONAR
    ) -> Optional[List[SelectionRange]]:
        ...

    @rpc_method(name="textDocument/selectionRange", param_type=SelectionRangeParams)
    async def _text_document_selection_range(
        self,
        text_document: TextDocumentIdentifier,
        positions: List[Position],
        *args: Any,
        **kwargs: Any,
    ) -> Optional[List[SelectionRange]]:

        results: List[SelectionRange] = []

        document = await self.parent.documents.get(text_document.uri)
        if document is None:
            return None

        for result in await self.collect(self, document, positions, callback_filter=language_id_filter(document)):
            if isinstance(result, BaseException):
                if not isinstance(result, CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:
                if result is not None:
                    results.extend(result)

        if len(results) == 0:
            return None

        return results



from __future__ import annotations

from asyncio import CancelledError
from typing import TYPE_CHECKING, Any, List, Optional

from ....jsonrpc2.protocol import rpc_method
from ....utils.async_tools import async_tasking_event
from ....utils.logging import LoggingDescriptor
from ..decorators import language_id_filter
from ..has_extend_capabilities import HasExtendCapabilities
from ..lsp_types import (
    DocumentFormattingOptions,
    DocumentFormattingParams,
    DocumentRangeFormattingOptions,
    DocumentRangeFormattingParams,
    FormattingOptions,
    ProgressToken,
    Range,
    ServerCapabilities,
    TextDocumentIdentifier,
    TextEdit,
)
from ..text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import LanguageServerProtocol

from .protocol_part import LanguageServerProtocolPart


class FormattingProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):

    _logger = LoggingDescriptor()

    def __init__(self, parent: LanguageServerProtocol) -> None:
        super().__init__(parent)

    @async_tasking_event
    async def format(
        sender, document: TextDocument, options: FormattingOptions, **further_options: Any  # NOSONAR
    ) -> Optional[List[TextEdit]]:
        ...

    @async_tasking_event
    async def format_range(
        sender, document: TextDocument, range: Range, options: FormattingOptions, **further_options: Any  # NOSONAR
    ) -> Optional[List[TextEdit]]:
        ...

    def extend_capabilities(self, capabilities: ServerCapabilities) -> None:
        if len(self.format):
            capabilities.document_formatting_provider = DocumentFormattingOptions(work_done_progress=True)
        if len(self.format_range):
            capabilities.document_range_formatting_provider = DocumentRangeFormattingOptions(work_done_progress=True)

    @rpc_method(name="textDocument/formatting", param_type=DocumentFormattingParams)
    async def _text_document_formatting(
        self,
        params: DocumentFormattingParams,
        text_document: TextDocumentIdentifier,
        options: FormattingOptions,
        work_done_token: Optional[ProgressToken],
        *args: Any,
        **kwargs: Any,
    ) -> Optional[List[TextEdit]]:

        results: List[TextEdit] = []

        document = await self.parent.documents.get(text_document.uri)
        if document is None:
            return None

        for result in await self.format(
            self,
            document,
            options,
            callback_filter=language_id_filter(document),
            **kwargs,
        ):
            if isinstance(result, BaseException):
                if not isinstance(result, CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:
                if result is not None:
                    results += result

        if len(results) > 0:
            return results

        return None

    @rpc_method(name="textDocument/rangeFormatting", param_type=DocumentRangeFormattingParams)
    async def _text_document_range_formatting(
        self,
        params: DocumentFormattingParams,
        text_document: TextDocumentIdentifier,
        range: Range,
        options: FormattingOptions,
        work_done_token: Optional[ProgressToken],
        *args: Any,
        **kwargs: Any,
    ) -> Optional[List[TextEdit]]:

        results: List[TextEdit] = []
        document = await self.parent.documents.get(text_document.uri)
        if document is None:
            return None

        for result in await self.format_range(
            self,
            document,
            range,
            options,
            callback_filter=language_id_filter(document),
            **kwargs,
        ):
            if isinstance(result, BaseException):
                if not isinstance(result, CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:
                if result is not None:
                    results += result

        if len(results) > 0:
            return results

        return None



from __future__ import annotations

import collections
import functools
import inspect
import logging
import os
import reprlib
import time
from enum import Enum
from typing import Any, Callable, List, Optional, Type, TypeVar, Union, cast, overload

__all__ = ["LoggingDescriptor"]


_my_repr: Optional[reprlib.Repr] = None


def get_repr() -> reprlib.Repr:
    global _my_repr
    if _my_repr is None:
        _my_repr = reprlib.Repr()
        _my_repr.maxother = 100
    return _my_repr


def _repr(o: Any) -> str:
    return get_repr().repr(o)
    # return repr(o)


def get_class_that_defined_method(meth: Callable[..., Any]) -> Optional[Type[Any]]:
    if inspect.ismethod(meth):
        for c in inspect.getmro(meth.__self__.__class__):
            if c.__dict__.get(meth.__name__) is meth:
                return c
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        class_name = meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]
        try:
            cls = getattr(inspect.getmodule(meth), class_name)
        except AttributeError:
            cls = meth.__globals__.get(class_name)
        if isinstance(cls, type):
            return cls
    return None


def _get_callable_has_self_or_cls_parameter(func: Any) -> bool:
    if inspect.ismethod(func) or isinstance(func, (classmethod, staticmethod)):
        return True
    if inspect.isfunction(func) and len(func.__qualname__.split(".<locals>.", 1)[-1].rsplit(".", 1)) > 1:
        return True
    return False


def get_unwrapped_func(func: Callable[..., Any]) -> Callable[..., Any]:
    result = inspect.unwrap(func)
    if isinstance(result, (staticmethod, classmethod)):
        return get_unwrapped_func(result.__func__)
    return cast(Callable[..., Any], result)


_LoggerEntry = collections.namedtuple("_LoggerEntry", "level prefix condition states")


class _HasLoggerEntries:
    __logging_entries__: List[_LoggerEntry]


_FUNC_TYPE = Union[Callable[[], logging.Logger], Callable[[], None], staticmethod, None]

_F = TypeVar("_F", bound=Callable[..., Any])


class LoggerException(Exception):
    pass


class CallState(Enum):
    ENTERING = "entering"
    EXITING = "exiting"
    EXCEPTION = "exception"


TRACE = logging.DEBUG - 6
logging.addLevelName(TRACE, "TRACE")


class LoggingDescriptor:
    __func: _FUNC_TYPE = None
    __name: Optional[str] = None
    __level: int = logging.NOTSET
    __owner: Any = None
    __owner_name: Optional[str] = None
    __postfix: str = ""
    __logger: Optional[logging.Logger] = None

    def __init__(
        self,
        _func: _FUNC_TYPE = None,
        *,
        name: Optional[str] = None,
        postfix: str = "",
        level: int = logging.NOTSET,
    ) -> None:
        self.__func = _func

        if _func is not None:
            functools.update_wrapper(self, _func)  # type: ignore

        self.__name = name
        self.__level = level
        self.__postfix = postfix

    def __init_logger(self) -> LoggingDescriptor:
        if self.__logger is None:
            returned_logger = None

            if self.__func is not None:

                if isinstance(self.__func, staticmethod):
                    returned_logger = self.__func.__func__()
                else:
                    returned_logger = self.__func()

            self.__logger = (
                returned_logger
                if returned_logger is not None
                else logging.getLogger(
                    self.__name
                    if self.__name is not None
                    else (
                        ("" if self.__owner is None else self.__owner.__module__ + "." + self.__owner.__qualname__)
                        if self.__owner is not None
                        else get_unwrapped_func(self.__func).__module__
                        if self.__func is not None
                        else "<unknown>"
                    )
                    + self.__postfix
                )
            )

            if self.__logger is None:
                raise NotImplementedError("Can't get or create a Logger object")

            self.set_level(self.__level)

        return self

    @property
    def logger(self) -> logging.Logger:
        if self.__logger is None:
            self.__init_logger()

        if self.__logger is None:
            raise LoggerException("Logger not initialized")

        return self.__logger

    def __set_name__(self, owner: Any, name: str) -> None:
        self.__owner = owner
        self.__owner_name = name

    def __call__(self, _func: _FUNC_TYPE = None) -> LoggingDescriptor:
        if _func is not None:
            self.__func = _func

        return self

    def __get__(self, obj: Any, objtype: Type[Any]) -> LoggingDescriptor:
        return self

    def log(
        self,
        level: int,
        msg: Any,
        condition: Optional[Callable[[], bool]] = None,
        *args: Any,
        stacklevel: int = 2,
        **kwargs: Any,
    ) -> None:
        if self.is_enabled_for(level) and condition is not None and condition() or condition is None:
            self.logger.log(level, msg() if callable(msg) else msg, *args, stacklevel=stacklevel, **kwargs)

    def debug(
        self,
        msg: Union[str, Callable[[], str]],
        condition: Optional[Callable[[], bool]] = None,
        *args: Any,
        stacklevel: int = 3,
        **kwargs: Any,
    ) -> None:
        self.log(logging.DEBUG, msg, condition, *args, stacklevel=stacklevel, **kwargs)

    def info(
        self,
        msg: Union[str, Callable[[], str]],
        condition: Optional[Callable[[], bool]] = None,
        *args: Any,
        stacklevel: int = 3,
        **kwargs: Any,
    ) -> None:
        self.log(logging.INFO, msg, condition, *args, stacklevel=stacklevel, **kwargs)

    def warning(
        self,
        msg: Union[str, Callable[[], str]],
        condition: Optional[Callable[[], bool]] = None,
        *args: Any,
        stacklevel: int = 3,
        **kwargs: Any,
    ) -> None:
        self.log(logging.WARNING, msg, condition, *args, stacklevel=stacklevel, **kwargs)

    def error(
        self,
        msg: Union[str, Callable[[], str]],
        condition: Optional[Callable[[], bool]] = None,
        *args: Any,
        stacklevel: int = 3,
        **kwargs: Any,
    ) -> None:
        self.log(logging.ERROR, msg, condition, *args, stacklevel=stacklevel, **kwargs)

    def trace(
        self,
        msg: Union[str, Callable[[], str]],
        condition: Optional[Callable[[], bool]] = None,
        *args: Any,
        stacklevel: int = 3,
        **kwargs: Any,
    ) -> None:
        self.log(TRACE, msg, condition, *args, stacklevel=stacklevel, **kwargs)

    def exception(
        self,
        msg: Union[BaseException, str, Callable[[], Union[BaseException, str]]],
        condition: Optional[Callable[[], bool]] = None,
        exc_info: Any = True,
        *args: Any,
        stacklevel: int = 3,
        level: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        if isinstance(msg, BaseException):
            sm = str(msg)
            s = type(msg).__qualname__
            if sm:
                s += ": " + sm
            self.log(
                logging.ERROR if level is None else level,
                s,
                condition,
                *args,
                exc_info=exc_info,
                stacklevel=stacklevel,
                **kwargs,
            )
        else:
            self.log(
                logging.ERROR if level is None else level,
                msg,
                condition,
                *args,
                exc_info=exc_info,
                stacklevel=stacklevel,
                **kwargs,
            )

    def critical(
        self,
        msg: Union[str, Callable[[], str]],
        condition: Optional[Callable[[], bool]] = None,
        *args: Any,
        stacklevel: int = 3,
        **kwargs: Any,
    ) -> None:
        self.log(logging.CRITICAL, msg, condition, *args, stacklevel=stacklevel, **kwargs)

    def is_enabled_for(self, level: int) -> bool:
        return self.logger.isEnabledFor(level)

    def set_level(self, level: int) -> None:
        self.logger.setLevel(level)

    def get_effective_level(self) -> int:
        return self.logger.getEffectiveLevel()

    def has_handlers(self) -> bool:
        return self.logger.hasHandlers()

    @property
    def name(self) -> str:
        return self.logger.name

    def __repr__(self) -> str:
        logger = self.logger
        level = logging.getLevelName(logger.getEffectiveLevel())
        return f"{self.__class__.__name__}(name={repr(logger.name)}, level={repr(level)})"

    _call_tracing_enabled = (
        "ROBOT_CALL_TRACING_ENABLED" in os.environ and os.environ["ROBOT_CALL_TRACING_ENABLED"] != "0"
    )
    _call_tracing_default_level = (
        logging.getLevelName(os.environ["ROBOT_CALL_TRACING_LEVEL"])
        if "ROBOT_CALL_TRACING_LEVEL" in os.environ
        else TRACE
    )

    @classmethod
    def set_call_tracing(cls, value: bool) -> None:
        cls._call_tracing_enabled = value

    @classmethod
    def set_call_tracing_default_level(cls, level: int) -> None:
        cls._call_tracing_default_level = level

    @overload
    def call(self, _func: _F) -> _F:
        ...

    @overload
    def call(
        self,
        *,
        level: Optional[int] = None,
        prefix: str = "",
        condition: Optional[Callable[..., bool]] = None,
        entering: bool = True,
        exiting: bool = False,
        exception: bool = False,
        timed: bool = False,
    ) -> Callable[[_F], _F]:
        ...

    def call(
        self,
        _func: Optional[_F] = None,
        *,
        level: Optional[int] = None,
        prefix: str = "",
        condition: Optional[Callable[..., bool]] = None,
        entering: bool = True,
        exiting: bool = False,
        exception: bool = False,
        timed: bool = False,
        **kwargs: Any,
    ) -> Callable[[_F], _F]:

        if level is None:
            level = type(self)._call_tracing_default_level

        def _decorator(func: _F) -> Callable[[_F], _F]:
            unwrapped_func = inspect.unwrap(func)

            if not hasattr(unwrapped_func, "__logging_entries__"):
                unwrapped_func.__logging_entries__ = []

            unwrapped_func.__logging_entries__.append(
                _LoggerEntry(
                    level=level,
                    prefix=prefix,
                    condition=condition,
                    states={CallState.ENTERING: entering, CallState.EXITING: exiting, CallState.EXCEPTION: exception},
                )
            )

            skip_first_arg = _get_callable_has_self_or_cls_parameter(unwrapped_func)

            @functools.wraps(func)
            def _wrapper(*wrapper_args: Any, **wrapper_kwargs: Any) -> Any:

                if isinstance(unwrapped_func, staticmethod):
                    real_args = wrapper_args[1:]
                    real_func = unwrapped_func.__func__
                else:
                    real_args = wrapper_args
                    real_func = unwrapped_func

                def has_logging_entries() -> bool:
                    return hasattr(unwrapped_func, "__logging_entries__")

                def func_name() -> str:
                    if isinstance(unwrapped_func, staticmethod):
                        return (
                            unwrapped_func.__func__.__qualname__
                            if self.__owner or self.__name is None
                            else unwrapped_func.__func__.__name__
                        )
                    else:
                        return (
                            str(unwrapped_func.__qualname__)
                            if self.__owner is None or self.__name
                            else str(unwrapped_func.__name__)
                        )

                def _log(
                    message: Callable[[], str],
                    *,
                    state: CallState,
                    log_level: Optional[int] = None,
                    stacklevel: int = 4,
                    **log_kwargs: Any,
                ) -> None:
                    if has_logging_entries():
                        for c in cast(_HasLoggerEntries, unwrapped_func).__logging_entries__:
                            if c.states[state]:

                                def state_msg() -> str:
                                    return (
                                        (str(state.value) + " ")
                                        if state != CallState.ENTERING or c.states[CallState.EXITING]
                                        else ""
                                    )

                                self.log(
                                    log_level if log_level is not None else c.level,
                                    lambda: f"{state_msg()}{prefix}{message()}",
                                    condition=lambda: c.condition is None or c.condition(*real_args, **wrapper_kwargs),
                                    stacklevel=stacklevel,
                                    **{**kwargs, **log_kwargs},
                                )

                def build_enter_message() -> str:
                    message_args = wrapper_args[1:] if skip_first_arg else wrapper_args

                    return "{0}({1}{2}{3})".format(
                        func_name(),
                        ", ".join(_repr(a) for a in message_args),
                        (", " if len(message_args) > 0 and len(wrapper_kwargs) > 0 else ""),
                        (
                            ", ".join(f"{str(k)}={_repr(v)}" for k, v in wrapper_kwargs.items())
                            if len(wrapper_kwargs) > 0
                            else ""
                        ),
                    )

                def build_exit_message(res: Any, duration: Optional[float]) -> str:
                    return "{0}(...) -> {1}{2}".format(
                        func_name(), _repr(res), f" duration: {duration}" if duration is not None else ""
                    )

                def build_exception_message(exception: BaseException) -> str:
                    return "{0}(...) -> {1}: {2}".format(func_name(), type(exception).__qualname__, exception)

                _log(build_enter_message, state=CallState.ENTERING)

                result = None
                try:
                    start_time: float = 0
                    end_time: float = 0
                    if timed:
                        start_time = time.time()

                    result = real_func(*real_args, **wrapper_kwargs)

                    if timed:
                        end_time = time.time()

                except BaseException as e:
                    ex = e
                    _log(
                        lambda: build_exception_message(ex),
                        log_level=logging.ERROR,
                        state=CallState.EXCEPTION,
                        exc_info=True,
                    )
                    raise
                else:
                    _log(
                        lambda: build_exit_message(result, end_time - start_time if timed else None),
                        state=CallState.EXITING,
                    )
                return result

            return _wrapper

        def _empty__decorator(func: _F) -> Callable[[_F], _F]:
            return func

        if _func is None:
            return cast(Callable[[_F], _F], _decorator if type(self)._call_tracing_enabled else _empty__decorator)

        return _decorator(_func) if type(self)._call_tracing_enabled else _empty__decorator(_func)



from __future__ import annotations

import inspect
import io
import weakref
from typing import Any, Awaitable, Callable, Dict, List, Optional, TypeVar, Union, cast

from ...utils.async_tools import Lock, create_sub_task
from ...utils.logging import LoggingDescriptor
from ...utils.uri import Uri
from .lsp_types import DocumentUri, Range


class InvalidRangeError(Exception):
    pass


_T = TypeVar("_T")


class CacheEntry:
    def __init__(self, data: Any = None) -> None:
        self.data = data
        self.lock = Lock()


class TextDocument:
    _logger = LoggingDescriptor()

    def __init__(
        self,
        document_uri: DocumentUri,
        text: str,
        language_id: Optional[str] = None,
        version: Optional[int] = None,
    ) -> None:
        super().__init__()

        self._lock = Lock()
        self._references: weakref.WeakSet[Any] = weakref.WeakSet()
        self.document_uri = document_uri
        self.uri = Uri(self.document_uri).normalized()
        self.language_id = language_id
        self._version = version
        self._text = text
        self._orig_text = text
        self._orig_version = version
        self._lines: Optional[List[str]] = None
        self._cache: Dict[weakref.ref[Any], CacheEntry] = {}
        self._data: weakref.WeakKeyDictionary[Any, Any] = weakref.WeakKeyDictionary()
        self.opened_in_editor = False

    @property
    def references(self) -> weakref.WeakSet[Any]:  # pragma: no cover
        return self._references

    @property
    def version(self) -> Optional[int]:
        return self._version

    def __str__(self) -> str:  # pragma: no cover
        return self.__repr__()

    def __repr__(self) -> str:  # pragma: no cover
        return (
            f"TextDocument(uri={repr(self.uri)}, "
            f"language_id={repr(self.language_id)}, "
            f"version={repr(self._version)}"
            f")"
        )

    async def text(self) -> str:
        async with self._lock:
            return self._text

    async def save(self, version: Optional[int], text: Optional[str]) -> None:
        await self.apply_full_change(version, text, save=True)

    async def revert(self, version: Optional[int]) -> bool:
        if self._orig_text != self._text or self._orig_version != self._version:
            await self.apply_full_change(version or self._orig_version, self._orig_text)
            return True
        return False

    @_logger.call
    async def apply_none_change(self) -> None:
        async with self._lock:
            self._lines = None
            self._invalidate_cache()

    @_logger.call
    async def apply_full_change(self, version: Optional[int], text: Optional[str], *, save: bool = False) -> None:
        async with self._lock:
            if version is not None:
                self._version = version
            if text is not None:
                self._text = text
                self._lines = None
            if save:
                self._orig_text = self._text
            self._invalidate_cache()

    @_logger.call
    async def apply_incremental_change(self, version: Optional[int], range: Range, text: str) -> None:
        async with self._lock:
            try:
                if version is not None:
                    self._version = version

                if range.start > range.end:
                    raise InvalidRangeError(f"Start position is greater then end position {range}.")

                lines = self.__get_lines()

                (start_line, start_col), (end_line, end_col) = range

                if start_line == len(lines):
                    self._text = self._text + text
                    return

                with io.StringIO() as new_text:
                    for i, line in enumerate(lines):
                        if i < start_line or i > end_line:
                            new_text.write(line)
                            continue

                        if i == start_line:
                            new_text.write(line[:start_col])
                            new_text.write(text)

                        if i == end_line:
                            new_text.write(line[end_col:])

                    self._text = new_text.getvalue()
            finally:
                self._lines = None
                self._invalidate_cache()

    def __get_lines(self) -> List[str]:
        if self._lines is None:
            self._lines = self._text.splitlines(True)

        return self._lines

    async def get_lines(self) -> List[str]:
        async with self._lock:
            return self.__get_lines()

    def _invalidate_cache(self) -> None:
        self._cache.clear()

    @_logger.call
    async def invalidate_cache(self) -> None:
        async with self._lock:
            self._invalidate_cache()

    def _invalidate_data(self) -> None:
        self._data.clear()

    @_logger.call
    async def invalidate_data(self) -> None:
        async with self._lock:
            self._invalidate_data()

    async def __remove_cache_entry_safe(self, _ref: Any) -> None:
        if _ref in self._cache:
            async with self._lock:
                if _ref in self._cache:
                    self._cache.pop(_ref)

    def __remove_cache_entry(self, ref: Any) -> None:

        create_sub_task(self.__remove_cache_entry_safe(ref))

    def __get_cache_reference(self, entry: Callable[..., Any], /, *, add_remove: bool = True) -> weakref.ref[Any]:

        if inspect.ismethod(entry):
            reference: weakref.ref[Any] = weakref.WeakMethod(entry, self.__remove_cache_entry if add_remove else None)
        else:
            reference = weakref.ref(entry, self.__remove_cache_entry if add_remove else None)

        return reference

    async def get_cache(
        self,
        entry: Union[Callable[[TextDocument], Awaitable[_T]], Callable[..., Awaitable[_T]]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:

        reference = self.__get_cache_reference(entry)
        e = self._cache.get(reference, None)

        if e is None:
            async with self._lock:
                e = self._cache.get(reference, None)
                if e is None:

                    e = CacheEntry()

                    self._cache[reference] = e

        if e.data is None:
            async with e.lock:
                if e.data is None:
                    e.data = await entry(self, *args, **kwargs)

        return cast("_T", e.data)

    @_logger.call
    async def remove_cache_entry(
        self, entry: Union[Callable[[TextDocument], Awaitable[_T]], Callable[..., Awaitable[_T]]]
    ) -> None:
        await self.__remove_cache_entry_safe(self.__get_cache_reference(entry, add_remove=False))

    def set_data(self, key: Any, data: Any) -> None:
        self._data[key] = data

    def get_data(self, key: Any, default: Optional[_T] = None) -> _T:
        return self._data.get(key, default)

    def _clear(self) -> None:
        self._lines = None
        self._invalidate_cache()
        self._invalidate_data()

    @_logger.call
    async def clear(self) -> None:
        async with self._lock:
            self._clear()



from typing import Protocol, runtime_checkable

from .lsp_types import ServerCapabilities

__all__ = ["HasExtendCapabilities"]


@runtime_checkable
class HasExtendCapabilities(Protocol):
    def extend_capabilities(self, capabilities: ServerCapabilities) -> None:
        ...



import argparse
import logging
import logging.config
import os
import pathlib
import sys
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any, Optional

__file__ = os.path.abspath(__file__)
if __file__.endswith((".pyc", ".pyo")):
    __file__ = __file__[:-1]

if __name__ == "__main__" and __package__ is None or __package__ == "":

    file = Path(__file__).resolve()
    parent, top = file.parent, file.parents[2]

    if str(top) not in sys.path:
        sys.path.append(str(top))

    try:
        sys.path.remove(str(parent))
    except ValueError:  # Already removed
        pass

    __package__ = "robotcode.language_server"

from .._version import __version__
from ..utils.debugpy import start_debugpy
from ..utils.logging import LoggingDescriptor

_logger = LoggingDescriptor(name=__package__)


def get_log_handler(logfile: str) -> logging.FileHandler:
    log_fn = pathlib.Path(logfile)
    roll_over = log_fn.exists()

    handler = RotatingFileHandler(log_fn, backupCount=5)
    formatter = logging.Formatter(
        fmt="[%(levelname)-7s] %(asctime)s (%(name)s) %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    handler.setFormatter(formatter)

    if roll_over:
        handler.doRollover()

    return handler


def run_server(mode: str, port: int, pipe_name: Optional[str]) -> int:
    from ..jsonrpc2.server import JsonRpcServerMode, TcpParams
    from .robotframework.server import RobotLanguageServer

    with RobotLanguageServer(
        mode=JsonRpcServerMode(mode), tcp_params=TcpParams("127.0.0.1", port), pipe_name=pipe_name
    ) as server:
        try:
            server.run()
        except SystemExit:
            raise
        except KeyboardInterrupt:
            return 1
        except BaseException as e:
            _logger.exception(e)
            return 1

        return 0


def create_parser() -> argparse.ArgumentParser:
    result = argparse.ArgumentParser(
        description="RobotCode Language Server",
        prog=__package__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    result.add_argument("--version", action="store_true", help="shows the version and exits")

    result.add_argument(
        "-m",
        "--mode",
        default="stdio",
        choices=["stdio", "pipe", "socket", "tcp"],
        help="communication mode",
    )

    result.add_argument("--stdio", action="store_true", help="run in stdio mode (Shortcut for --mode=stdio)")
    result.add_argument(
        "--socket", default=None, help="run in socket mode (Shortcut for --mode=socket --port ...)", type=int
    )
    result.add_argument(
        "--pipe", default=None, help="run in named pipe mode (Shortcut for --mode=pipe --pipe-name ...)", type=str
    )

    result.add_argument("-p", "--port", default=6610, help="server listen port (mode socket and tcp)", type=int)

    result.add_argument("-pn", "--pipe-name", default=None, help="pipe name for pipe mode", type=str)

    result.add_argument("--log", action="store_true", help="enable logging")
    result.add_argument("--log-json-rpc", action="store_true", help="show json-rpc log messages")
    result.add_argument("--log-json-rpc-data", action="store_true", help="show json-rpc-data log messages")
    result.add_argument("--log-language-server", action="store_true", help="show language server log messages")
    result.add_argument(
        "--log-language-server-parts", action="store_true", help="show language server parts log messages"
    )
    result.add_argument(
        "--log-robotframework", action="store_true", help="show robotframework language server log messages"
    )
    result.add_argument("--debug-asyncio", action="store_true", help="enable async io debugging messages")
    result.add_argument("--log-asyncio", action="store_true", help="show asyncio log messages")
    result.add_argument("--log-colored", action="store_true", help="colored output for logs")
    result.add_argument("--log-config", default=None, help="reads logging configuration from file", metavar="FILE")
    result.add_argument("--log-file", default=None, help="enables logging to file", metavar="FILE")
    result.add_argument("--log-level", default="WARNING", help="sets the overall log level", metavar="LEVEL")
    result.add_argument("--call-tracing", action="store_true", help="enables log tracing of method calls")
    result.add_argument(
        "--call-tracing-default-level", default="TRACE", help="sets the default level for call tracing", metavar="LEVEL"
    )
    result.add_argument("--debugpy", action="store_true", help="starts a debugpy session")
    result.add_argument(
        "--debugpy-port", default=5678, help="sets the port for debugpy session", type=int, metavar="PORT"
    )
    result.add_argument("--debugpy-wait-for-client", action="store_true", help="waits for debugpy client to connect")
    return result


def init_colored_logs(log_level: Any) -> bool:
    try:
        import coloredlogs

        coloredlogs.install(level=log_level)
    except ImportError:
        return False

    return True


def init_logging(args: argparse.Namespace) -> None:
    log_level = logging._checkLevel(args.log_level) if args.log else logging.WARNING  # type: ignore

    log_initialized = False
    if args.log_colored:
        log_initialized = init_colored_logs(log_level)

    if not log_initialized:
        logging.basicConfig(level=log_level, format="%(name)s:%(levelname)s: %(message)s")

    if args.log_file is not None:
        _logger.logger.addHandler(get_log_handler(args.log_file))

    if not args.log_robotframework:
        logging.getLogger("robotcode.language_server.robotframework").setLevel(logging.CRITICAL)
    if not args.log_language_server_parts:
        logging.getLogger("robotcode.language_server.common.parts").setLevel(logging.CRITICAL)
    if not args.log_language_server:
        logging.getLogger("robotcode.language_server.common").setLevel(logging.CRITICAL)
    if not args.log_json_rpc:
        logging.getLogger("robotcode.jsonrpc2").setLevel(logging.CRITICAL)
    if not args.log_json_rpc_data:
        logging.getLogger("robotcode.jsonrpc2.protocol.JsonRPCProtocol_data").setLevel(logging.CRITICAL)
    if not args.log_asyncio:
        logging.getLogger("asyncio").setLevel(logging.CRITICAL)


def main() -> None:
    parser = create_parser()

    args = parser.parse_args()

    if args.version:
        print(__version__)
        return

    if args.call_tracing:
        LoggingDescriptor.set_call_tracing(True)
    if args.call_tracing_default_level:
        LoggingDescriptor.set_call_tracing_default_level(
            logging._checkLevel(args.call_tracing_default_level)  # type: ignore
        )

    if args.debug_asyncio:
        os.environ["PYTHONASYNCIODEBUG"] = "1"
        logging.getLogger("asyncio").level = logging.DEBUG
    else:
        logging.getLogger("asyncio").level = logging.CRITICAL

    if args.log:
        if args.log_config is not None:
            if not os.path.exists(args.log_config):
                raise FileNotFoundError(f"Log-config file '{args.log_config}' not exists.")

            logging.config.fileConfig(args.log_config, disable_existing_loggers=True)
        else:
            init_logging(args)

    _logger.info(f"starting language server version={__version__}")
    _logger.debug(f"args={args}")

    if args.debugpy:
        start_debugpy(args.debugpy_port, args.debugpy_wait_for_client)

    mode = args.mode

    if args.stdio:
        mode = "stdio"

    port = args.port

    if args.socket is not None:
        port = args.socket
        mode = "socket"

    pipe_name = args.pipe_name
    if args.pipe is not None:
        mode = "pipe"
        pipe_name = args.pipe

    sys.exit(run_server(mode, port, pipe_name))


if __name__ == "__main__":
    main()



import ast
from typing import Any, AsyncGenerator, Callable, Optional, Type, cast

__all__ = ["iter_fields", "iter_child_nodes", "AsyncVisitor", "walk"]


async def iter_fields(node: ast.AST) -> AsyncGenerator[Any, None]:
    """
    Yield a tuple of ``(fieldname, value)`` for each field in ``node._fields``
    that is present on *node*.
    """
    for field in node._fields:
        try:
            yield field, getattr(node, field)
        except AttributeError:
            pass


async def iter_child_nodes(node: ast.AST) -> AsyncGenerator[ast.AST, None]:
    """
    Yield all direct child nodes of *node*, that is, all fields that are nodes
    and all items of fields that are lists of nodes.
    """
    async for _name, field in iter_fields(node):
        if isinstance(field, ast.AST):
            yield field
        elif isinstance(field, list):
            for item in field:
                if isinstance(item, ast.AST):
                    yield item


async def walk(node: ast.AST) -> AsyncGenerator[ast.AST, None]:
    from collections import deque

    todo = deque([node])
    while todo:
        node = todo.popleft()
        todo.extend([e async for e in iter_child_nodes(node)])
        yield node


async def iter_nodes(node: ast.AST) -> AsyncGenerator[ast.AST, None]:
    async for _name, value in iter_fields(node):
        if isinstance(value, list):
            for item in value:
                if isinstance(item, ast.AST):
                    yield item
                    async for n in iter_nodes(item):
                        yield n

        elif isinstance(value, ast.AST):
            yield value

            async for n in iter_nodes(value):
                yield n


class VisitorFinder:
    def _find_visitor(self, cls: Type[Any]) -> Optional[Callable[..., Any]]:
        if cls is ast.AST:
            return None
        method_name = "visit_" + cls.__name__
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            if callable(method):
                return cast("Callable[..., Any]", method)
        for base in cls.__bases__:
            method = self._find_visitor(base)
            if method:
                return cast("Callable[..., Any]", method)
        return None


class AsyncVisitor(VisitorFinder):
    async def visit(self, node: ast.AST) -> None:
        visitor = self._find_visitor(type(node)) or self.generic_visit
        await visitor(node)

    async def generic_visit(self, node: ast.AST) -> None:
        """Called if no explicit visitor function exists for a node."""
        async for field, value in iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        await self.visit(item)
            elif isinstance(value, ast.AST):
                await self.visit(value)



from __future__ import annotations

import ast
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from ....utils.async_itertools import async_next
from ....utils.async_tools import threaded
from ....utils.logging import LoggingDescriptor
from ...common.decorators import language_id
from ...common.lsp_types import DocumentHighlight, DocumentHighlightKind, Position
from ...common.text_document import TextDocument
from ..diagnostics.entities import VariableDefinition
from ..utils.ast_utils import (
    HasTokens,
    Statement,
    Token,
    get_nodes_at_position,
    get_tokens_at_position,
    range_from_token,
)

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol

from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart

_DocumentHighlightMethod = Callable[[ast.AST, TextDocument, Position], Awaitable[Optional[List[DocumentHighlight]]]]


class RobotDocumentHighlightProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.document_highlight.collect.add(self.collect)

    def _find_method(self, cls: Type[Any]) -> Optional[_DocumentHighlightMethod]:
        if cls is ast.AST:
            return None
        method_name = "highlight_" + cls.__name__
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            if callable(method):
                return cast(_DocumentHighlightMethod, method)
        for base in cls.__bases__:
            method = self._find_method(base)
            if method:
                return cast(_DocumentHighlightMethod, method)
        return None

    @language_id("robotframework")
    @threaded()
    @_logger.call
    async def collect(
        self,
        sender: Any,
        document: TextDocument,
        position: Position,
    ) -> Optional[List[DocumentHighlight]]:

        result_nodes = await get_nodes_at_position(await self.parent.documents_cache.get_model(document), position)

        if not result_nodes:
            return None

        result_node = result_nodes[-1]

        if result_node is None:
            return None

        result = await self._highlight_default(result_nodes, document, position)
        if result:
            return result

        method = self._find_method(type(result_node))
        if method is not None:
            result = await method(result_node, document, position)
            if result is not None:
                return result

        return None

    async def _highlight_default(
        self, nodes: List[ast.AST], document: TextDocument, position: Position
    ) -> Optional[List[DocumentHighlight]]:
        from robot.parsing.lexer.tokens import Token as RobotToken

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        if not nodes:
            return None

        node = nodes[-1]

        if not isinstance(node, HasTokens):
            return None

        tokens = get_tokens_at_position(node, position)

        token_and_var: Optional[Tuple[Token, VariableDefinition]] = None

        for token in tokens:
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_variables_from_token(token, namespace, nodes, position)
                    if position in range_from_token(var_token)
                ),
                None,
            )

        if (
            token_and_var is None
            and isinstance(node, Statement)
            and isinstance(node, self.get_expression_statement_types())
            and (token := node.get_token(RobotToken.ARGUMENT)) is not None
            and position in range_from_token(token)
        ):
            token_and_var = await async_next(
                (
                    (var_token, var)
                    async for var_token, var in self.iter_expression_variables_from_token(
                        token, namespace, nodes, position
                    )
                    if position in range_from_token(var_token)
                ),
                None,
            )

        if token_and_var is not None:
            _, variable = token_and_var

            return [
                DocumentHighlight(e.range, DocumentHighlightKind.TEXT)
                for e in await self.parent.robot_references.find_variable_references_in_file(document, variable)
            ]

        return None

    async def highlight_KeywordCall(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position
    ) -> Optional[List[DocumentHighlight]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordCall

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordCall, node)
        result = await self.get_keyworddoc_and_token_from_position(
            kw_node.keyword,
            cast(Token, kw_node.get_token(RobotToken.KEYWORD)),
            [cast(Token, t) for t in kw_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    # TODO highlight namespaces
                    return None
            if (
                position in kw_range
                and keyword_doc is not None
                and not keyword_doc.is_error_handler
                and keyword_doc.source
            ):
                return [
                    *(
                        [DocumentHighlight(keyword_doc.range, DocumentHighlightKind.TEXT)]
                        if keyword_doc.source == str(document.uri.to_path())
                        else []
                    ),
                    *(
                        DocumentHighlight(e.range, DocumentHighlightKind.TEXT)
                        for e in await self.parent.robot_references.find_keyword_references_in_file(
                            document, keyword_doc
                        )
                    ),
                ]

        return None

    async def highlight_KeywordName(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position
    ) -> Optional[List[DocumentHighlight]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import KeywordName

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        kw_node = cast(KeywordName, node)

        name_token = cast(RobotToken, kw_node.get_token(RobotToken.KEYWORD_NAME))

        if not name_token:
            return None

        doc = await namespace.get_library_doc()
        if doc is not None:
            keyword = next(
                (v for v in doc.keywords.keywords if v.name == name_token.value and v.line_no == kw_node.lineno),
                None,
            )

            if keyword is not None and keyword.source and not keyword.is_error_handler:
                return [
                    DocumentHighlight(keyword.range, DocumentHighlightKind.TEXT),
                    *(
                        DocumentHighlight(e.range, DocumentHighlightKind.TEXT)
                        for e in await self.parent.robot_references.find_keyword_references_in_file(document, keyword)
                    ),
                ]

        return None

    async def highlight_Fixture(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position
    ) -> Optional[List[DocumentHighlight]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Fixture

        namespace = await self.parent.documents_cache.get_namespace(document)
        if namespace is None:
            return None

        fixture_node = cast(Fixture, node)

        name_token = cast(Token, fixture_node.get_token(RobotToken.NAME))
        if name_token is None or name_token.value is None or name_token.value.upper() in ("", "NONE"):
            return None

        result = await self.get_keyworddoc_and_token_from_position(
            fixture_node.name,
            name_token,
            [cast(Token, t) for t in fixture_node.get_tokens(RobotToken.ARGUMENT)],
            namespace,
            position,
        )

        if result is not None:
            keyword_doc, keyword_token = result

            keyword_token = self.strip_bdd_prefix(keyword_token)

            lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

            kw_range = range_from_token(keyword_token)

            if lib_entry and kw_namespace:
                r = range_from_token(keyword_token)
                r.end.character = r.start.character + len(kw_namespace)
                kw_range.start.character = r.end.character + 1
                if position in r:
                    # TODO highlight namespaces
                    return None

            if position in kw_range and keyword_doc is not None and not keyword_doc.is_error_handler:
                return [
                    *(
                        [DocumentHighlight(keyword_doc.range, DocumentHighlightKind.TEXT)]
                        if keyword_doc.source == str(document.uri.to_path())
                        else []
                    ),
                    *(
                        DocumentHighlight(e.range, DocumentHighlightKind.TEXT)
                        for e in await self.parent.robot_references.find_keyword_references_in_file(
                            document, keyword_doc
                        )
                    ),
                ]

        return None

    async def _highlight_Template_or_TestTemplate(  # noqa: N802
        self, node: ast.AST, document: TextDocument, position: Position
    ) -> Optional[List[DocumentHighlight]]:
        from robot.parsing.lexer.tokens import Token as RobotToken
        from robot.parsing.model.statements import Template, TestTemplate

        template_node = cast(Union[Template, TestTemplate], node)
        if template_node.value:

            keyword_token = cast(RobotToken, template_node.get_token(RobotToken.NAME))
            if keyword_token is None or keyword_token.value is None or keyword_token.value.upper() in ("", "NONE"):
                return None

            keyword_token = self.strip_bdd_prefix(keyword_token)

            if position.is_in_range(range_from_token(keyword_token), False):
                namespace = await self.parent.documents_cache.get_namespace(document)
                if namespace is None:
                    return None

                keyword_doc = await namespace.find_keyword(template_node.value)

                if keyword_doc is not None:

                    lib_entry, kw_namespace = await self.get_namespace_info_from_keyword(namespace, keyword_token)

                    kw_range = range_from_token(keyword_token)

                    if lib_entry and kw_namespace:
                        r = range_from_token(keyword_token)
                        r.end.character = r.start.character + len(kw_namespace)
                        kw_range.start.character = r.end.character + 1
                        if position in r:
                            # TODO highlight namespaces
                            return None

                    if not keyword_doc.is_error_handler:
                        return [
                            *(
                                [DocumentHighlight(keyword_doc.range, DocumentHighlightKind.TEXT)]
                                if keyword_doc.source == str(document.uri.to_path())
                                else []
                            ),
                            *(
                                DocumentHighlight(e.range, DocumentHighlightKind.TEXT)
                                for e in await self.parent.robot_references.find_keyword_references_in_file(
                                    document, keyword_doc
                                )
                            ),
                        ]
        return None

    async def highlight_TestTemplate(  # noqa: N802
        self, result_node: ast.AST, document: TextDocument, position: Position
    ) -> Optional[List[DocumentHighlight]]:
        return await self._highlight_Template_or_TestTemplate(result_node, document, position)

    async def highlight_Template(  # noqa: N802
        self, result_node: ast.AST, document: TextDocument, position: Position
    ) -> Optional[List[DocumentHighlight]]:
        return await self._highlight_Template_or_TestTemplate(result_node, document, position)



from __future__ import annotations

import asyncio
import inspect
import json
import threading
import traceback
from collections import OrderedDict
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from ..jsonrpc2.protocol import (
    JsonRPCException,
    JsonRPCProtocolBase,
    SendedRequestEntry,
)
from ..utils.dataclasses import as_dict, as_json, from_dict
from ..utils.inspect import ensure_coroutine
from ..utils.logging import LoggingDescriptor
from .dap_types import (
    ErrorBody,
    ErrorResponse,
    Event,
    Message,
    ProtocolMessage,
    Request,
    Response,
)


class DebugAdapterErrorResponseError(JsonRPCException):
    def __init__(
        self,
        error: ErrorResponse,
    ) -> None:
        super().__init__(
            f'{error.message} (seq={error.request_seq} command="{error.command}")'
            f'{f": {error.body.error}" if error.body is not None and error.body.error  else ""}',
        )
        self.error = error


class DebugAdapterRPCErrorException(JsonRPCException):
    def __init__(
        self,
        message: Optional[str] = None,
        request_seq: int = -1,
        command: str = "",
        success: Optional[bool] = None,
        error_message: Optional[Message] = None,
    ) -> None:
        super().__init__(
            f'{(message+" ") if message else ""}(seq={request_seq} command="{command}")'
            f'{f": {error_message}" if error_message else ""}'
        )
        self.message = message
        self.request_seq = request_seq
        self.command = command
        self.success = success
        self.error_message = error_message


TResult = TypeVar("TResult", bound=Any)


class DebugAdapterProtocol(JsonRPCProtocolBase):

    _logger = LoggingDescriptor()

    def __init__(self) -> None:
        super().__init__()
        self._sended_request_lock = threading.RLock()
        self._sended_request: OrderedDict[int, SendedRequestEntry] = OrderedDict()
        self._received_request_lock = threading.RLock()
        self._received_request: OrderedDict[int, asyncio.Future[Any]] = OrderedDict()
        self._initialized = False

    @_logger.call
    def send_message(self, message: ProtocolMessage) -> None:
        body = as_json(message, compact=True).encode(self.CHARSET)

        header = (f"Content-Length: {len(body)}\r\n\r\n").encode("ascii")

        if self.write_transport is not None:
            msg = header + body

            if self._loop:
                self.write_transport.write(msg)

    def send_error(
        self,
        message: Optional[str] = None,
        request_seq: int = -1,
        command: str = "",
        success: Optional[bool] = None,
        error_message: Optional[Message] = None,
    ) -> None:
        self.send_message(
            ErrorResponse(
                success=success or False,
                request_seq=request_seq,
                message=message,
                command=command,
                body=ErrorBody(error=error_message),
            )
        )

    @staticmethod
    def _generate_json_rpc_messages_from_dict(
        data: Union[Dict[Any, Any], List[Dict[Any, Any]]]
    ) -> Iterator[ProtocolMessage]:
        def inner(d: Dict[Any, Any]) -> ProtocolMessage:
            result = from_dict(d, (Request, Response, Event))
            if isinstance(result, Response) and not result.success:
                result = from_dict(d, ErrorResponse)
            return result

        if isinstance(data, list):
            for e in data:
                yield inner(e)
        else:
            yield inner(data)

    def _handle_body(self, body: bytes, charset: str) -> None:
        try:
            self._handle_messages(self._generate_json_rpc_messages_from_dict(json.loads(body.decode(charset))))
        except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
            raise
        except BaseException as e:
            self._logger.exception(e)
            self.send_error(
                f"Invalid Message: {type(e).__name__}: {str(e)} -> {str(body)}\n{traceback.format_exc()}",
                error_message=Message(traceback.format_exc()),
            )

    def _handle_messages(self, iterator: Iterator[ProtocolMessage]) -> None:
        def done(f: asyncio.Future[Any]) -> None:
            ex = f.exception()
            if ex is not None and not isinstance(ex, asyncio.CancelledError):
                self._logger.exception(ex, exc_info=ex)

        for m in iterator:
            task = asyncio.create_task(self.handle_message(m))
            task.add_done_callback(done)

    @_logger.call
    async def handle_message(self, message: ProtocolMessage) -> None:
        if isinstance(message, Request):
            self.handle_request(message)
        if isinstance(message, Event):
            self.handle_event(message)
        elif isinstance(message, ErrorResponse):
            self.handle_error_response(message)
        elif isinstance(message, Response):
            self.handle_response(message)

    @staticmethod
    def _convert_params(
        callable: Callable[..., Any], param_type: Optional[Type[Any]], params: Any
    ) -> Tuple[List[Any], Dict[str, Any]]:
        if params is None:
            return [], {}
        if param_type is None:
            if isinstance(params, Mapping):
                return [], dict(**params)
            else:
                return [params], {}

        converted_params = from_dict(params, param_type)

        signature = inspect.signature(callable)

        has_var_kw = any(p.kind == inspect.Parameter.VAR_KEYWORD for p in signature.parameters.values())

        kw_args = {}
        args = []
        params_added = False
        rest = set(converted_params.__dict__.keys())
        if isinstance(params, dict):
            rest = set.union(rest, params.keys())

        for v in signature.parameters.values():
            if v.name in converted_params.__dict__:
                if v.kind == inspect.Parameter.POSITIONAL_ONLY:
                    args.append(getattr(converted_params, v.name))
                else:
                    kw_args[v.name] = getattr(converted_params, v.name)
                rest.remove(v.name)
            elif v.name == "arguments":
                if v.kind == inspect.Parameter.POSITIONAL_ONLY:
                    args.append(converted_params)
                    params_added = True
                else:
                    kw_args[v.name] = converted_params
                    params_added = True
            elif isinstance(params, dict) and v.name in params:
                if v.kind == inspect.Parameter.POSITIONAL_ONLY:
                    args.append(params[v.name])
                else:
                    kw_args[v.name] = params[v.name]
        if has_var_kw:
            for r in rest:
                if hasattr(converted_params, r):
                    kw_args[r] = getattr(converted_params, r)
                elif isinstance(params, dict) and r in params:
                    kw_args[r] = params[r]

            if not params_added:
                kw_args["arguments"] = converted_params
        return args, kw_args

    async def handle_unknown_command(self, message: Request) -> Any:
        raise DebugAdapterRPCErrorException(
            f"Unknown Command '{message.command}'",
            error_message=Message(
                format='Unknown command "{command}"', variables={"command": str(message.command)}, show_user=True
            ),
        )

    @_logger.call
    def handle_request(self, message: Request) -> None:
        e = self.registry.get_entry(message.command)

        with self._received_request_lock:
            if e is None or not callable(e.method):
                result = asyncio.create_task(self.handle_unknown_command(message))
            else:
                params = self._convert_params(e.method, e.param_type, message.arguments)

                result = asyncio.create_task(ensure_coroutine(e.method)(*params[0], **params[1]))

            self._received_request[message.seq] = result

        def done(t: asyncio.Task[Any]) -> None:
            try:
                self.send_response(message.seq, message.command, t.result())
            except asyncio.CancelledError:
                self._logger.info(f"request message {repr(message)} canceled")
            except (SystemExit, KeyboardInterrupt):
                raise
            except DebugAdapterRPCErrorException as ex:
                self._logger.exception(ex)
                self.send_error(
                    message=ex.message,
                    request_seq=message.seq,
                    command=ex.command or message.command,
                    success=ex.success or False,
                    error_message=ex.error_message,
                )
            except DebugAdapterErrorResponseError as ex:
                self.send_error(
                    ex.error.message,
                    message.seq,
                    message.command,
                    False,
                    error_message=ex.error.body.error if ex.error.body is not None else None,
                )
            except BaseException as e:
                self._logger.exception(e)
                self.send_error(
                    str(type(e).__name__),
                    message.seq,
                    message.command,
                    False,
                    error_message=Message(format=f"{type(e).__name__}: {e}", show_user=True),
                )
            finally:
                with self._received_request_lock:
                    self._received_request.pop(message.seq, None)

        result.add_done_callback(done)

    @_logger.call
    def send_response(
        self,
        request_seq: int,
        command: str,
        result: Optional[Any] = None,
        success: bool = True,
        message: Optional[str] = None,
    ) -> None:
        self.send_message(
            Response(request_seq=request_seq, command=command, success=success, message=message, body=result)
        )

    def send_request(
        self,
        request: Request,
        return_type: Optional[Type[TResult]] = None,
    ) -> asyncio.Future[TResult]:

        result: asyncio.Future[TResult] = asyncio.get_event_loop().create_future()

        with self._sended_request_lock:
            self._sended_request[request.seq] = SendedRequestEntry(result, return_type)

        self.send_message(request)

        return result

    async def send_request_async(
        self,
        request: Request,
        return_type: Optional[Type[TResult]] = None,
    ) -> TResult:
        return await self.send_request(request, return_type)

    @_logger.call
    def send_event(self, event: Event) -> None:
        self.send_message(event)

    @_logger.call
    async def send_event_async(self, event: Event) -> None:
        self.send_event(event)

    @_logger.call
    def handle_error_response(self, message: ErrorResponse) -> None:
        with self._sended_request_lock:
            entry = self._sended_request.pop(message.request_seq, None)

        exception = DebugAdapterErrorResponseError(message)
        if entry is None:
            raise exception

        entry.future.set_exception(exception)

    @_logger.call
    def handle_response(self, message: Response) -> None:
        with self._sended_request_lock:
            entry = self._sended_request.pop(message.request_seq, None)

        if entry is None:
            error = f"Invalid response. Could not find id '{message.request_seq}' in request list {message!r}"
            self._logger.warning(error)
            self.send_error("invalid response", error_message=Message(format=error, show_user=True))
            return

        try:
            if message.success:
                if not entry.future.done():
                    entry.future.set_result(from_dict(message.body, entry.result_type))
            else:
                raise DebugAdapterErrorResponseError(ErrorResponse(**as_dict(message)))
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as e:
            if not entry.future.done():
                entry.future.set_exception(e)

    @_logger.call
    def handle_event(self, message: Event) -> None:
        raise NotImplementedError()



from __future__ import annotations

from asyncio import CancelledError
from typing import TYPE_CHECKING, Any, List, Optional

from ....jsonrpc2.protocol import rpc_method
from ....utils.async_tools import async_tasking_event
from ....utils.logging import LoggingDescriptor
from ..decorators import language_id_filter
from ..has_extend_capabilities import HasExtendCapabilities
from ..lsp_types import (
    FoldingRange,
    FoldingRangeParams,
    ServerCapabilities,
    TextDocumentIdentifier,
)
from ..text_document import TextDocument

if TYPE_CHECKING:
    from ..protocol import LanguageServerProtocol  # pragma: no cover

from .protocol_part import LanguageServerProtocolPart


class FoldingRangeProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):

    _logger = LoggingDescriptor()

    def __init__(self, parent: LanguageServerProtocol) -> None:
        super().__init__(parent)

    @async_tasking_event
    async def collect(sender, document: TextDocument) -> Optional[List[FoldingRange]]:  # pragma: no cover, NOSONAR
        ...

    def extend_capabilities(self, capabilities: ServerCapabilities) -> None:
        if len(self.collect):
            capabilities.folding_range_provider = True

    @rpc_method(name="textDocument/foldingRange", param_type=FoldingRangeParams)
    async def _text_document_folding_range(
        self, text_document: TextDocumentIdentifier, *args: Any, **kwargs: Any
    ) -> Optional[List[FoldingRange]]:

        results: List[FoldingRange] = []
        document = await self.parent.documents.get(text_document.uri)
        if document is None:
            return None

        for result in await self.collect(self, document, callback_filter=language_id_filter(document)):
            if isinstance(result, BaseException):
                if not isinstance(result, CancelledError):
                    self._logger.exception(result, exc_info=result)
            else:
                if result is not None:
                    results += result

        if len(results) == 0:
            return None
        return results
